/*
 * Copyright 2023 Intel Corporation
 * SPDX-License-Identifier: MIT
 */

/* This is an automatically generated file. */

#ifndef _BUILTIN_TYPES_
#define _BUILTIN_TYPES_

extern const struct glsl_type glsl_type_builtin_error;
extern const struct glsl_type glsl_type_builtin_void;
extern const struct glsl_type glsl_type_builtin_bool;
extern const struct glsl_type glsl_type_builtin_bvec2;
extern const struct glsl_type glsl_type_builtin_bvec3;
extern const struct glsl_type glsl_type_builtin_bvec4;
extern const struct glsl_type glsl_type_builtin_bvec5;
extern const struct glsl_type glsl_type_builtin_bvec8;
extern const struct glsl_type glsl_type_builtin_bvec16;
extern const struct glsl_type glsl_type_builtin_int;
extern const struct glsl_type glsl_type_builtin_ivec2;
extern const struct glsl_type glsl_type_builtin_ivec3;
extern const struct glsl_type glsl_type_builtin_ivec4;
extern const struct glsl_type glsl_type_builtin_ivec5;
extern const struct glsl_type glsl_type_builtin_ivec8;
extern const struct glsl_type glsl_type_builtin_ivec16;
extern const struct glsl_type glsl_type_builtin_uint;
extern const struct glsl_type glsl_type_builtin_uvec2;
extern const struct glsl_type glsl_type_builtin_uvec3;
extern const struct glsl_type glsl_type_builtin_uvec4;
extern const struct glsl_type glsl_type_builtin_uvec5;
extern const struct glsl_type glsl_type_builtin_uvec8;
extern const struct glsl_type glsl_type_builtin_uvec16;
extern const struct glsl_type glsl_type_builtin_float;
extern const struct glsl_type glsl_type_builtin_vec2;
extern const struct glsl_type glsl_type_builtin_vec3;
extern const struct glsl_type glsl_type_builtin_vec4;
extern const struct glsl_type glsl_type_builtin_vec5;
extern const struct glsl_type glsl_type_builtin_vec8;
extern const struct glsl_type glsl_type_builtin_vec16;
extern const struct glsl_type glsl_type_builtin_float16_t;
extern const struct glsl_type glsl_type_builtin_f16vec2;
extern const struct glsl_type glsl_type_builtin_f16vec3;
extern const struct glsl_type glsl_type_builtin_f16vec4;
extern const struct glsl_type glsl_type_builtin_f16vec5;
extern const struct glsl_type glsl_type_builtin_f16vec8;
extern const struct glsl_type glsl_type_builtin_f16vec16;
extern const struct glsl_type glsl_type_builtin_double;
extern const struct glsl_type glsl_type_builtin_dvec2;
extern const struct glsl_type glsl_type_builtin_dvec3;
extern const struct glsl_type glsl_type_builtin_dvec4;
extern const struct glsl_type glsl_type_builtin_dvec5;
extern const struct glsl_type glsl_type_builtin_dvec8;
extern const struct glsl_type glsl_type_builtin_dvec16;
extern const struct glsl_type glsl_type_builtin_int64_t;
extern const struct glsl_type glsl_type_builtin_i64vec2;
extern const struct glsl_type glsl_type_builtin_i64vec3;
extern const struct glsl_type glsl_type_builtin_i64vec4;
extern const struct glsl_type glsl_type_builtin_i64vec5;
extern const struct glsl_type glsl_type_builtin_i64vec8;
extern const struct glsl_type glsl_type_builtin_i64vec16;
extern const struct glsl_type glsl_type_builtin_uint64_t;
extern const struct glsl_type glsl_type_builtin_u64vec2;
extern const struct glsl_type glsl_type_builtin_u64vec3;
extern const struct glsl_type glsl_type_builtin_u64vec4;
extern const struct glsl_type glsl_type_builtin_u64vec5;
extern const struct glsl_type glsl_type_builtin_u64vec8;
extern const struct glsl_type glsl_type_builtin_u64vec16;
extern const struct glsl_type glsl_type_builtin_int16_t;
extern const struct glsl_type glsl_type_builtin_i16vec2;
extern const struct glsl_type glsl_type_builtin_i16vec3;
extern const struct glsl_type glsl_type_builtin_i16vec4;
extern const struct glsl_type glsl_type_builtin_i16vec5;
extern const struct glsl_type glsl_type_builtin_i16vec8;
extern const struct glsl_type glsl_type_builtin_i16vec16;
extern const struct glsl_type glsl_type_builtin_uint16_t;
extern const struct glsl_type glsl_type_builtin_u16vec2;
extern const struct glsl_type glsl_type_builtin_u16vec3;
extern const struct glsl_type glsl_type_builtin_u16vec4;
extern const struct glsl_type glsl_type_builtin_u16vec5;
extern const struct glsl_type glsl_type_builtin_u16vec8;
extern const struct glsl_type glsl_type_builtin_u16vec16;
extern const struct glsl_type glsl_type_builtin_int8_t;
extern const struct glsl_type glsl_type_builtin_i8vec2;
extern const struct glsl_type glsl_type_builtin_i8vec3;
extern const struct glsl_type glsl_type_builtin_i8vec4;
extern const struct glsl_type glsl_type_builtin_i8vec5;
extern const struct glsl_type glsl_type_builtin_i8vec8;
extern const struct glsl_type glsl_type_builtin_i8vec16;
extern const struct glsl_type glsl_type_builtin_uint8_t;
extern const struct glsl_type glsl_type_builtin_u8vec2;
extern const struct glsl_type glsl_type_builtin_u8vec3;
extern const struct glsl_type glsl_type_builtin_u8vec4;
extern const struct glsl_type glsl_type_builtin_u8vec5;
extern const struct glsl_type glsl_type_builtin_u8vec8;
extern const struct glsl_type glsl_type_builtin_u8vec16;
extern const struct glsl_type glsl_type_builtin_mat2;
extern const struct glsl_type glsl_type_builtin_mat3;
extern const struct glsl_type glsl_type_builtin_mat4;
extern const struct glsl_type glsl_type_builtin_mat2x3;
extern const struct glsl_type glsl_type_builtin_mat2x4;
extern const struct glsl_type glsl_type_builtin_mat3x2;
extern const struct glsl_type glsl_type_builtin_mat3x4;
extern const struct glsl_type glsl_type_builtin_mat4x2;
extern const struct glsl_type glsl_type_builtin_mat4x3;
extern const struct glsl_type glsl_type_builtin_f16mat2;
extern const struct glsl_type glsl_type_builtin_f16mat3;
extern const struct glsl_type glsl_type_builtin_f16mat4;
extern const struct glsl_type glsl_type_builtin_f16mat2x3;
extern const struct glsl_type glsl_type_builtin_f16mat2x4;
extern const struct glsl_type glsl_type_builtin_f16mat3x2;
extern const struct glsl_type glsl_type_builtin_f16mat3x4;
extern const struct glsl_type glsl_type_builtin_f16mat4x2;
extern const struct glsl_type glsl_type_builtin_f16mat4x3;
extern const struct glsl_type glsl_type_builtin_dmat2;
extern const struct glsl_type glsl_type_builtin_dmat3;
extern const struct glsl_type glsl_type_builtin_dmat4;
extern const struct glsl_type glsl_type_builtin_dmat2x3;
extern const struct glsl_type glsl_type_builtin_dmat2x4;
extern const struct glsl_type glsl_type_builtin_dmat3x2;
extern const struct glsl_type glsl_type_builtin_dmat3x4;
extern const struct glsl_type glsl_type_builtin_dmat4x2;
extern const struct glsl_type glsl_type_builtin_dmat4x3;
extern const struct glsl_type glsl_type_builtin_atomic_uint;
extern const struct glsl_type glsl_type_builtin_sampler;
extern const struct glsl_type glsl_type_builtin_sampler1D;
extern const struct glsl_type glsl_type_builtin_sampler2D;
extern const struct glsl_type glsl_type_builtin_sampler3D;
extern const struct glsl_type glsl_type_builtin_samplerCube;
extern const struct glsl_type glsl_type_builtin_sampler1DArray;
extern const struct glsl_type glsl_type_builtin_sampler2DArray;
extern const struct glsl_type glsl_type_builtin_samplerCubeArray;
extern const struct glsl_type glsl_type_builtin_sampler2DRect;
extern const struct glsl_type glsl_type_builtin_samplerBuffer;
extern const struct glsl_type glsl_type_builtin_sampler2DMS;
extern const struct glsl_type glsl_type_builtin_sampler2DMSArray;
extern const struct glsl_type glsl_type_builtin_isampler1D;
extern const struct glsl_type glsl_type_builtin_isampler2D;
extern const struct glsl_type glsl_type_builtin_isampler3D;
extern const struct glsl_type glsl_type_builtin_isamplerCube;
extern const struct glsl_type glsl_type_builtin_isampler1DArray;
extern const struct glsl_type glsl_type_builtin_isampler2DArray;
extern const struct glsl_type glsl_type_builtin_isamplerCubeArray;
extern const struct glsl_type glsl_type_builtin_isampler2DRect;
extern const struct glsl_type glsl_type_builtin_isamplerBuffer;
extern const struct glsl_type glsl_type_builtin_isampler2DMS;
extern const struct glsl_type glsl_type_builtin_isampler2DMSArray;
extern const struct glsl_type glsl_type_builtin_usampler1D;
extern const struct glsl_type glsl_type_builtin_usampler2D;
extern const struct glsl_type glsl_type_builtin_usampler3D;
extern const struct glsl_type glsl_type_builtin_usamplerCube;
extern const struct glsl_type glsl_type_builtin_usampler1DArray;
extern const struct glsl_type glsl_type_builtin_usampler2DArray;
extern const struct glsl_type glsl_type_builtin_usamplerCubeArray;
extern const struct glsl_type glsl_type_builtin_usampler2DRect;
extern const struct glsl_type glsl_type_builtin_usamplerBuffer;
extern const struct glsl_type glsl_type_builtin_usampler2DMS;
extern const struct glsl_type glsl_type_builtin_usampler2DMSArray;
extern const struct glsl_type glsl_type_builtin_samplerShadow;
extern const struct glsl_type glsl_type_builtin_sampler1DShadow;
extern const struct glsl_type glsl_type_builtin_sampler2DShadow;
extern const struct glsl_type glsl_type_builtin_samplerCubeShadow;
extern const struct glsl_type glsl_type_builtin_sampler1DArrayShadow;
extern const struct glsl_type glsl_type_builtin_sampler2DArrayShadow;
extern const struct glsl_type glsl_type_builtin_samplerCubeArrayShadow;
extern const struct glsl_type glsl_type_builtin_sampler2DRectShadow;
extern const struct glsl_type glsl_type_builtin_samplerExternalOES;
extern const struct glsl_type glsl_type_builtin_texture1D;
extern const struct glsl_type glsl_type_builtin_texture2D;
extern const struct glsl_type glsl_type_builtin_texture3D;
extern const struct glsl_type glsl_type_builtin_textureCube;
extern const struct glsl_type glsl_type_builtin_texture1DArray;
extern const struct glsl_type glsl_type_builtin_texture2DArray;
extern const struct glsl_type glsl_type_builtin_textureCubeArray;
extern const struct glsl_type glsl_type_builtin_texture2DRect;
extern const struct glsl_type glsl_type_builtin_textureBuffer;
extern const struct glsl_type glsl_type_builtin_texture2DMS;
extern const struct glsl_type glsl_type_builtin_texture2DMSArray;
extern const struct glsl_type glsl_type_builtin_itexture1D;
extern const struct glsl_type glsl_type_builtin_itexture2D;
extern const struct glsl_type glsl_type_builtin_itexture3D;
extern const struct glsl_type glsl_type_builtin_itextureCube;
extern const struct glsl_type glsl_type_builtin_itexture1DArray;
extern const struct glsl_type glsl_type_builtin_itexture2DArray;
extern const struct glsl_type glsl_type_builtin_itextureCubeArray;
extern const struct glsl_type glsl_type_builtin_itexture2DRect;
extern const struct glsl_type glsl_type_builtin_itextureBuffer;
extern const struct glsl_type glsl_type_builtin_itexture2DMS;
extern const struct glsl_type glsl_type_builtin_itexture2DMSArray;
extern const struct glsl_type glsl_type_builtin_utexture1D;
extern const struct glsl_type glsl_type_builtin_utexture2D;
extern const struct glsl_type glsl_type_builtin_utexture3D;
extern const struct glsl_type glsl_type_builtin_utextureCube;
extern const struct glsl_type glsl_type_builtin_utexture1DArray;
extern const struct glsl_type glsl_type_builtin_utexture2DArray;
extern const struct glsl_type glsl_type_builtin_utextureCubeArray;
extern const struct glsl_type glsl_type_builtin_utexture2DRect;
extern const struct glsl_type glsl_type_builtin_utextureBuffer;
extern const struct glsl_type glsl_type_builtin_utexture2DMS;
extern const struct glsl_type glsl_type_builtin_utexture2DMSArray;
extern const struct glsl_type glsl_type_builtin_textureExternalOES;
extern const struct glsl_type glsl_type_builtin_vtexture1D;
extern const struct glsl_type glsl_type_builtin_vtexture2D;
extern const struct glsl_type glsl_type_builtin_vtexture3D;
extern const struct glsl_type glsl_type_builtin_vtexture1DArray;
extern const struct glsl_type glsl_type_builtin_vtexture2DArray;
extern const struct glsl_type glsl_type_builtin_vtextureBuffer;
extern const struct glsl_type glsl_type_builtin_image1D;
extern const struct glsl_type glsl_type_builtin_image2D;
extern const struct glsl_type glsl_type_builtin_image3D;
extern const struct glsl_type glsl_type_builtin_image2DRect;
extern const struct glsl_type glsl_type_builtin_imageCube;
extern const struct glsl_type glsl_type_builtin_imageBuffer;
extern const struct glsl_type glsl_type_builtin_image1DArray;
extern const struct glsl_type glsl_type_builtin_image2DArray;
extern const struct glsl_type glsl_type_builtin_imageCubeArray;
extern const struct glsl_type glsl_type_builtin_image2DMS;
extern const struct glsl_type glsl_type_builtin_image2DMSArray;
extern const struct glsl_type glsl_type_builtin_iimage1D;
extern const struct glsl_type glsl_type_builtin_iimage2D;
extern const struct glsl_type glsl_type_builtin_iimage3D;
extern const struct glsl_type glsl_type_builtin_iimage2DRect;
extern const struct glsl_type glsl_type_builtin_iimageCube;
extern const struct glsl_type glsl_type_builtin_iimageBuffer;
extern const struct glsl_type glsl_type_builtin_iimage1DArray;
extern const struct glsl_type glsl_type_builtin_iimage2DArray;
extern const struct glsl_type glsl_type_builtin_iimageCubeArray;
extern const struct glsl_type glsl_type_builtin_iimage2DMS;
extern const struct glsl_type glsl_type_builtin_iimage2DMSArray;
extern const struct glsl_type glsl_type_builtin_uimage1D;
extern const struct glsl_type glsl_type_builtin_uimage2D;
extern const struct glsl_type glsl_type_builtin_uimage3D;
extern const struct glsl_type glsl_type_builtin_uimage2DRect;
extern const struct glsl_type glsl_type_builtin_uimageCube;
extern const struct glsl_type glsl_type_builtin_uimageBuffer;
extern const struct glsl_type glsl_type_builtin_uimage1DArray;
extern const struct glsl_type glsl_type_builtin_uimage2DArray;
extern const struct glsl_type glsl_type_builtin_uimageCubeArray;
extern const struct glsl_type glsl_type_builtin_uimage2DMS;
extern const struct glsl_type glsl_type_builtin_uimage2DMSArray;
extern const struct glsl_type glsl_type_builtin_i64image1D;
extern const struct glsl_type glsl_type_builtin_i64image2D;
extern const struct glsl_type glsl_type_builtin_i64image3D;
extern const struct glsl_type glsl_type_builtin_i64image2DRect;
extern const struct glsl_type glsl_type_builtin_i64imageCube;
extern const struct glsl_type glsl_type_builtin_i64imageBuffer;
extern const struct glsl_type glsl_type_builtin_i64image1DArray;
extern const struct glsl_type glsl_type_builtin_i64image2DArray;
extern const struct glsl_type glsl_type_builtin_i64imageCubeArray;
extern const struct glsl_type glsl_type_builtin_i64image2DMS;
extern const struct glsl_type glsl_type_builtin_i64image2DMSArray;
extern const struct glsl_type glsl_type_builtin_u64image1D;
extern const struct glsl_type glsl_type_builtin_u64image2D;
extern const struct glsl_type glsl_type_builtin_u64image3D;
extern const struct glsl_type glsl_type_builtin_u64image2DRect;
extern const struct glsl_type glsl_type_builtin_u64imageCube;
extern const struct glsl_type glsl_type_builtin_u64imageBuffer;
extern const struct glsl_type glsl_type_builtin_u64image1DArray;
extern const struct glsl_type glsl_type_builtin_u64image2DArray;
extern const struct glsl_type glsl_type_builtin_u64imageCubeArray;
extern const struct glsl_type glsl_type_builtin_u64image2DMS;
extern const struct glsl_type glsl_type_builtin_u64image2DMSArray;
extern const struct glsl_type glsl_type_builtin_vbuffer;
extern const struct glsl_type glsl_type_builtin_vimage1D;
extern const struct glsl_type glsl_type_builtin_vimage2D;
extern const struct glsl_type glsl_type_builtin_vimage3D;
extern const struct glsl_type glsl_type_builtin_vimage1DArray;
extern const struct glsl_type glsl_type_builtin_vimage2DArray;
extern const struct glsl_type glsl_type_builtin_subpassInput;
extern const struct glsl_type glsl_type_builtin_subpassInputMS;
extern const struct glsl_type glsl_type_builtin_isubpassInput;
extern const struct glsl_type glsl_type_builtin_isubpassInputMS;
extern const struct glsl_type glsl_type_builtin_usubpassInput;
extern const struct glsl_type glsl_type_builtin_usubpassInputMS;
extern const struct glsl_type glsl_type_builtin_textureSubpassInput;
extern const struct glsl_type glsl_type_builtin_textureSubpassInputMS;
extern const struct glsl_type glsl_type_builtin_itextureSubpassInput;
extern const struct glsl_type glsl_type_builtin_itextureSubpassInputMS;
extern const struct glsl_type glsl_type_builtin_utextureSubpassInput;
extern const struct glsl_type glsl_type_builtin_utextureSubpassInputMS;

#endif /* _BUILTIN_TYPES_ */