/*
 * Copyright © 2011 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/**
 * \file string_to_uint_map.cpp
 * \brief Dumb wrapprs so that C code can create and destroy maps.
 *
 * \author Ian Romanick <ian.d.romanick@intel.com>
 */
#include "string_to_uint_map.h"

extern "C" struct string_to_uint_map *
string_to_uint_map_ctor()
{
   return new string_to_uint_map;
}

extern "C" void
string_to_uint_map_dtor(struct string_to_uint_map *map)
{
   delete map;
}

extern "C" void
string_to_uint_map_put(struct string_to_uint_map *map,
                       unsigned value, const char *key)
{
   map->put(value, key);
}

extern "C" bool
string_to_uint_map_get(struct string_to_uint_map *map,
                       unsigned *value, const char *key)
{
   unsigned val;
   bool r = map->get(val, key);
   if (r)
      *value = val;
   return r;
}
